
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <dirent.h>
#include <time.h>
#include <sys/stat.h>

#define ANSWERS_REQUIRED 64
char *key = "None hold a candle to Bessie the sh33p!";
char *secret = "illogical";

int numFiles;
char **files;
unsigned int *fileVals;

void sigchld(int sig) {
   int status;
   while (wait4(-1, &status, WNOHANG, NULL) > 0) {}
}

int read_line(int fd, char *buf, unsigned int size) {
   char ch;
   unsigned int total = 0;
   while (1) {
      if (read(fd, &ch, 1) <= 0) {
         return -1;
      }
      if (ch == '\n') break;
      if (total >= size) return -1;
      buf[total++] = ch;
   }
   return (int)total;
}

int send_buf(int fd, unsigned char *buf, unsigned int size) {
   unsigned int total = 0;
   while (total < size) {
      int nbytes = send(fd, buf + total, size - total, 0);
      if (nbytes == 0) return -1;
      total += nbytes;
   }
   return (int)total;
}

void cleanup() {
   int i;
   for (i = 0; i < numFiles; i++) {
      free(files[i]);
   }
   free(files);
   free(fileVals);
}

int send_photo(char *fname, int fd) {
   struct stat sb;
   unsigned char buf[1024];
   if (!stat(fname, &sb)) {
      int f = open(fname, O_RDONLY);
      int len, i;
      off_t tot = 0;
      unsigned int sz = sb.st_size;
      if (f < 0) {
         fprintf(stderr, "Failed to open %s\n", fname);
         return 0;
      }
      sz = htonl(sz ^ 0xc932f768);
      send_buf(fd, (unsigned char*)&sz, sizeof(unsigned int));
      sz = sb.st_size & 0xFF;
      while (tot < sb.st_size) {
         size_t need = sb.st_size - tot;
         size_t req = (need > sizeof(buf)) ? sizeof(buf) : need;
         unsigned int bytes = read(f, buf, req);
         if (bytes == 0 || bytes > req) break;
         tot += bytes;
         for (i = 0; i < bytes; i++) {
            buf[i] ^= sz;
         }
         if (send_buf(fd, buf, bytes) != bytes) break;
      }
      close(f);
      return tot == sb.st_size;
   }
   else {
      fprintf(stderr, "Failed to stat %s\n", fname);
   }
   return 0;
}

int handle_client(int client_fd) {
   int r = open("/dev/urandom", O_RDONLY);
   unsigned int f1, f2;
   unsigned char choice;
   unsigned int correct = 0;
   unsigned int res;
   int len;
   struct timeval timeout = {20, 0}; 
   char answers[ANSWERS_REQUIRED];
   char userAnswers[ANSWERS_REQUIRED];
   char pass[64];
   if (r < 0) {
      fprintf(stderr, "failed to open /dev/urandom\n");
      return 1;
   }
   len = read_line(client_fd, pass, sizeof(pass) - 1);
   if (len == -1) {
      return 1;
   }
   pass[len] = 0;
   if (strcmp(pass, secret)) {
      return 1;
   }
   while (correct < ANSWERS_REQUIRED) {
      read(r, &f1, sizeof(unsigned int));
      f1 = f1 % numFiles;
      do {
         read(r, &f2, sizeof(unsigned int));
         f2 = f2 % numFiles;
      } while (f2 == f1);
      
      send_photo(files[f1], client_fd);
      send_photo(files[f2], client_fd);
      if (recv(client_fd, &choice, 1, 0) != 1) {
         close(r);
         res = htonl(0xffffffff ^ 0xc932f768);
         send_buf(client_fd, (unsigned char*)&res, sizeof(res));
         return 1;
      }
      if (choice > 1) {
         close(r);
         res = htonl(0xfffffffe ^ 0xc932f768);
         send_buf(client_fd, (unsigned char*)&res, sizeof(res));
         return 1;
      }
      if ((fileVals[f1] < fileVals[f2]) && choice == 0 ||
          (fileVals[f2] < fileVals[f1]) && choice == 1) {
         close(r);
         res = htonl(0xfffffffe ^ 0xc932f768);
         send_buf(client_fd, (unsigned char*)&res, sizeof(res));
         return 1;
      }
      answers[correct++] = choice == 0 ? 'L' : 'R';
   }
   close(r);

   res = htonl(0xfffffffd ^ 0xc932f768);
   send_buf(client_fd, (unsigned char*)&res, sizeof(res));
   if (setsockopt(client_fd, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout)) == -1) {
      err(-1, "Unable to set rcvtimeo");
   }
   len = recv(client_fd, &userAnswers, ANSWERS_REQUIRED, 0);
   if (len != ANSWERS_REQUIRED || 
       strncmp(userAnswers, answers, ANSWERS_REQUIRED)) {
      if (len == -1) {
         res = htonl(0xfffffffa ^ 0xc932f768);
      }
      else {
         res = htonl(0xfffffffb ^ 0xc932f768);
      }
      send_buf(client_fd, (unsigned char*)&res, sizeof(res));
      return 1;
   }

   res = htonl(0xfffffffc ^ 0xc932f768);
   send_buf(client_fd, (unsigned char*)&res, sizeof(res));
   res = htonl(strlen(key));
   send_buf(client_fd, (unsigned char*)&res, sizeof(res));
   send_buf(client_fd, key, strlen(key));
   return 0;
}

unsigned int computeRank(char *fname) {
   FILE *f = fopen(fname, "rb");
   unsigned int res = 0;
   unsigned int v = 0;
   while (fread(&v, sizeof(unsigned int), 1, f) == 1) {
      res = (res >> 13) || (res << 19);
      res ^= v;
   }
   fclose(f);
   return v;
}

void merge(int low, int mid, int high) {
   int i = low;
   int j = mid;
   int k = 0;
   char **cout = (char**)malloc((high - low) * sizeof(char*));
   unsigned int *iout = (unsigned int*)malloc((high - low) * sizeof(unsigned int));
   while (i < mid && j < high) {
      int idx = i;
      if (fileVals[i] <= fileVals[j]) {
         i++;
      }
      else {
         idx = j++;
      }
      cout[k] = files[idx];
      iout[k++] = fileVals[idx];
   }
   while (i < mid) {
      cout[k] = files[i];
      iout[k++] = fileVals[i++];
   }
   while (j < high) {
      cout[k] = files[j];
      iout[k++] = fileVals[j++];
   }
   k = 0;
   while (low < high) {
      files[low] = cout[k];
      fileVals[low++] = iout[k++];
   }
   free(cout);
   free(iout);
}

void sortFiles(int low, int high) {
   int diff = high - low;
   if (diff <= 1) return;
   int mid = (high + low) / 2;
   sortFiles(low, mid);
   sortFiles(mid, high);
   merge(low, mid, high);
}

int loadPhotoData(char *path) {
   struct dirent dir;
   int i;
   DIR *d = opendir(path);
   if (d) {
      while (1) {
         char *fname;
         struct stat sb;
         struct dirent *de = readdir(d);
         if (de == NULL) break;
         dir = *de;
         fname = (char*)malloc(strlen(path) + strlen(dir.d_name) + 2);
         sprintf(fname, "%s/%s", path, dir.d_name);
         if (stat(fname, &sb) == -1) {
            free(fname);
            continue;
         }
         char *ext = strstr(fname, ".jpg");
         if (S_ISREG(sb.st_mode) && ext && ext[4] == 0) {
            numFiles++;
            files = (char**)realloc(files, numFiles * sizeof(char*));
            files[numFiles - 1] = fname;
            fileVals = (unsigned int *)realloc(fileVals, sizeof(unsigned int) * numFiles);
            fileVals[numFiles - 1] = computeRank(fname);
         }
      }
      closedir(d);
      sortFiles(0, numFiles);
      return 1;
   }
   else {
      return 0;
   }
}

int main(int argc, char **argv, char **envp) {
   int fd;
   struct sockaddr_in sa;
   struct sigaction act;
   int one = 1;
   if (argc != 2) {
      fprintf(stderr, "Missing photo dir, quitting.\n");
      exit(1);
   }
   if (loadPhotoData(argv[1]) == 0) {
      fprintf(stderr, "Failed to load photo data, quitting.\n");
      exit(1);
   }
   memset(&sa, 0, sizeof(sa));
   sa.sin_family = AF_INET;
   sa.sin_port = htons(8129);
   signal(SIGCHLD, sigchld);
   fd = socket(AF_INET, SOCK_STREAM, 0);
   setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
   bind(fd, (struct sockaddr*)&sa, sizeof(sa));
   listen(fd, 20);
   while (1) {
      unsigned int sa_len = sizeof(sa);
      int client = accept(fd, (struct sockaddr*)&sa, &sa_len);
      if (client != -1) {
         if (!fork()) {
            close(fd);
            int result = handle_client(client);
            cleanup();
            close(client);
            exit(result);
         }
         close(client);
      }   
   }
   return 0;
}

